/* vim:set ts=4 sw=4:
 *
 * (C) 2008 Willem Hengeveld  itsme@xs4all.nl
 *
 * IdcPerl - a perl scripting plugin for the Interactive Disassembler by hex-rays.
 * see http://www.xs4all.nl/~itsme/projects/idcperl
 *
 * this file contains the code to init, and call the perl interpreter
 *
 */

// perl includes
#include "extern.h"
#include "perl.h"
#include "xsub.h"
#include "perlapi.h"    // PL_exit_flags

#ifdef DYNAMIC_PERL
#include "redefperlasdll.h"
#include "perldllprocs.h"
#endif

#include "perlinterp.h"
#include "perl_magic.h"
#include "iolayer.h"

#include "kernwin.hpp"  // for msg()
#include "diskio.hpp"   // for idadir()

#ifdef TRACE_INTERP
#define tracemsg msg
#else
#define tracemsg  while(0)
#endif
Perl::Perl(PerlInterpreter *interp/*=NULL*/) : perl_interp(interp)
{
    tracemsg("new Perl(%08lx, host=%08lx)\n", interp, interp?*(long*)interp:0x12345678);
}
Perl* Perl::clone()
{
    tracemsg("Perl::clone, cloning %08lx, host=%08lx\n", perl_interp, *(long*)perl_interp);
    PERL_SET_CONTEXT(perl_interp);
    return new Perl(perl_clone(perl_interp, CLONEf_CLONE_HOST));
}
bool Perl::initialize()
{
    char	*bootargs[] = { "IDC", NULL };
    static char *args[] = { "", "-e", "0" };

    tracemsg("Perl::initialize\n");
    if (perl_interp)
        return true;

    perl_interp = perl_alloc();
    if (perl_interp==NULL)
        return false;
    perl_construct(perl_interp);

    init_idamsg_io(aTHX);

    tracemsg("calling xs_init\n");
    int status= perl_parse(perl_interp, Perl::xs_init, sizeof(args)/sizeof(char*), args, 0);
    if (status)
        return false;

    // add perl subdir of the ida application dir to perl search path @INC
    // NOTE: need to do this -after- perl_parse, since PL_incgv is 
    // initialized in perl_parse
    av_push(GvAV(PL_incgv), newSVpv(idadir("perl"),0));

    PL_exit_flags |= PERL_EXIT_DESTRUCT_END;

    tracemsg("calling IDC::bootstrap\n");
    // 'use IDC'
    call_argv("IDC::bootstrap", (long)G_DISCARD, bootargs);

    IDC_init();   // add magic

//    require("IDC");


    tracemsg("Perl::initialize done\n");
    return true;
}

void Perl::free()
{
    tracemsg("Perl::free(%08lx)\n", perl_interp);
    if (perl_interp)
    {
        PERL_SET_CONTEXT(perl_interp);
        //perl_run(perl_interp); ... not needed because of PERL_EXIT_DESTRUCT_END
        perl_destruct(perl_interp);
        perl_free(perl_interp);
        perl_interp = NULL;
    }
}

void Perl::exec(const char*perlcode)
{
    PERL_SET_CONTEXT(perl_interp);

    SV *codesv= newSVpv(perlcode, 0);
    eval_sv(codesv, G_KEEPERR);

    // check eval result.
    if (SvTRUE(ERRSV))
        msg("IDAPERL ERROR: %s\n", SvPV_nolen(ERRSV));
}
void Perl::require(const char*modulename)
{
    tracemsg("Perl::require(%s)\n", modulename);
    PERL_SET_CONTEXT(perl_interp);
    SV *codesv= newSV(strlen(modulename)+10);
    sv_catpv(codesv, "require ");
    sv_catpv(codesv, modulename);

    eval_sv(codesv, G_KEEPERR);
    // check eval result.
    if (SvTRUE(ERRSV))
        msg("IDAPERL ERROR: %s\n", SvPV_nolen(ERRSV));
}

#ifndef DYNAMIC_PERL
EXTERN_C void (*boot_DynaLoader)_((pTHX_ CV*));
#endif
EXTERN_C void boot_IDC (pTHX_ CV*);

void Perl::xs_init(pTHX)
{
    tracemsg("Perl::xs_init\n");
    char *file = "idc_perl.xs";

#ifdef __NT__
    // dynaloader is needed to load perlXX.dll under windows.
    newXS("DynaLoader::boot_DynaLoader", boot_DynaLoader, file);
#endif
    newXS("IDC::bootstrap", boot_IDC, file);

    tracemsg("xs_init done\n");
}
